package it.polimi.deib.se.numbers;

/**
 * Represents a complex number.
 * 
 * @author Andrea Mocci
 * @version 1.0
 * @since 1.0
 */
public final class Complex {

	/* parte reale */
	private final double re;
	/* parte immaginaria */
	private final double im;
	
	/**
	 * Builds the zero.
	 */
	public Complex() {
		this(0.0d, 0.0d);
	}
	
	/**
	 * Constructs a generic complex number.
	 * 
	 * The constructor takes two parameters
	 * specifying the real and imaginary part of
	 * the complex number.
	 * @param re the real part
	 * @param im the imaginary part
	 * @throws IllegalArgumentException if either <code>re</code> or <code>re</code> are infinite or NaN.
	 */
	public Complex(double re, double im) {
		if (Double.isInfinite(re) || 
				Double.isInfinite(im))
			throw new IllegalArgumentException("la parte reale e immaginaria non possono essere infinite");
		if (Double.isNaN(re) || 
				Double.isNaN(im))
			throw new IllegalArgumentException("la parte reale e immaginaria non possono essere NaN");
		
		this.re = re;
		this.im = im;
	}
	
	/**
	 * Builds a complex number starting from its polar form.
	 * 
     * The polar form is specified with a non-negative absolute 
     * value and a phase in radiants
 	 * @param abs the non-negative absolute value of the complex number.
	 * @param phase the phase of the complex number in radiants.
	 * @return a complex number with the specified absolute value and phase.
	 * @throws IllegalArgumentException if 
	 * <code>abs</code> is negative, or if either <code>abs</code> or <code>phase</code> are NaN or infinite.
	 */
	public static Complex fromPolarForm(double abs, 
			double phase) {
		
		if (Double.isInfinite(abs) ||
				Double.isInfinite(phase))
			throw new IllegalArgumentException("abs and phase cannot be infinite");
		if (Double.isNaN(abs) ||
				Double.isNaN(phase))
			throw new IllegalArgumentException("abs and phase cannot be NaN");			
		
		if (abs < 0) {
			throw new IllegalArgumentException("abs cannot be negative");
		}
		
		/* ricorda : z = modulus * (cos(phase) + i sin(phase)) */		
		double realPart = abs * Math.cos(phase);
		double imaginaryPart = abs * Math.sin(phase);
		return new Complex(realPart, imaginaryPart);
	}
	
	/**
	 * Returns the real part of the complex number.
	 * @return the real part of the complex number.
	 */
	public double re() {
		return this.re;
	}
	
	/**
	 * Returns the imaginary part of the complex number.
	 * @return the imaginary part of the complex number.
	 */
	public double im() {
		return im;
	}
	

	/**
	 * Returns the absolute value of the complex number.
	 * @return the non-negative absolute value of the complex number.
	 */
	public double abs() {
		// siccome la classe e' immutabile, potremmo precomputarlo.
		return Math.sqrt(Math.pow(this.re, 2.0d) + Math.pow(this.im, 2.0d));
	}
	
	/**
	 * Returns the phase of the complex number, in radiants.
	 * @return the phase of the complex number.
	 */
	public double phase() {
		// a bit of spaghetti code...
		if (this.re != 0) {
			double tmp = Math.atan(this.im / this.re); 
			if (this.re < 0 && this.im >= 0) tmp += Math.PI;
			if (this.re < 0 && this.im < 0) tmp += -Math.PI;
			return tmp;
		} else { /* this.re == 0 */
			if (this.im < 0) return -Math.PI;
			/* if (this.im >= 0) (default) */
			return Math.PI;
		}
		// TODO: Cercate nella documentazione di java.lang.Math un modo per
		// implementare questo metodo in modo piu' conciso .
	}
	
	/**
	 * Returns the complex conjugate of <code>this</code>.
	 * @return the complex number of this.
	 */
	public Complex conjugate() {
		return new Complex(this.re, -this.im);
	}
	
	/**
	 * Returns the sum of this and another complex number.
	 * @param other a complex number
	 * @return the sum of <code>this</code> and <code>other</code>.
	 * @throws IllegalArgumentException if <code>other</code> is <code>null</code>.
	 */
	public Complex sum(Complex other) {
		if (other == null)
			throw new IllegalArgumentException("other cannot be null");

		return new Complex(this.re + other.re(),
				this.im + other.im());
	}
	
	/**
	 * Returns the difference between this and another complex numbers.
	 * @param other a complex number
	 * @return the difference between <code>this</code> and <code>other</code>.
	 * 	 * @throws IllegalArgumentException if <code>other</code> is <code>null</code>.
	 */
	public Complex diff(Complex other) {
		if (other == null)
			throw new IllegalArgumentException("other cannot be null");

		return new Complex(this.re - other.re(),
				this.im -  other.im());
	}
	
	/**
	 * Returns the product between this and another complex number.
	 * @param other a complex number.
	 * @return the product between <code>this</code> and <code>other</code>.
	 * @throws IllegalArgumentException if <code>other</code> is <code>null</code>.
	 */
	public Complex mult(Complex other) {
		if (other == null)
			throw new IllegalArgumentException("other cannot be null");
		double productRe = this.re() * other.re() - this.im() * other.im();
		double productIm = this.re() * other.im() + this.im() * other.re();
		return new Complex(productRe, productIm);
	}
	
	/**
	 * Checks that <code>this</code> is approximately equal to
	 * another complex number.
	 * 
	 * Formally, approximate equality means that 
	 * the difference between the absolute value of 
	 *  <code>this</code> and the absolute value of 
	 * <code>other</code> is less than <code>precision</code>.
	 * @param other the complex number compared to 
	 *   <code>this</code>
	 * @param precision the difference, in absolute value,
	 * after which the two complex numbers are considered
	 * different.
	 * @return <code>true</code> if the two complex
	 * numbers are equal, <code>false</code>
	 * otherwise.
	 * @throws IllegalArgumentException if <code>precision</code>
	 * is either negative, infinite or NaN.
	 * @throws IllegalArgumentException if <code>other</code>
	 * is <code>null</code>.
	 */
	public boolean approximatelyEquals(Complex other, 
			double precision) {
		if (other == null)
			throw new IllegalArgumentException("other cannot be null");
		if (precision < 0.0d ||
				Double.isNaN(precision) ||
				Double.isInfinite(precision))
			throw new IllegalArgumentException("precision cannot be negative, NaN, or infinite");
		return (this.diff(other).abs() < precision);
	}


	/**
	 * Returns the roots of the complex number with a 
	 * certain degree
	 * 
	 * @see ComplexSet
	 * @param degree the degree of the roots to be calculated.
	 * @return a {@link ComplexSet} that contains the roots of 
	 * the complex number. The <code>size()</code> of the set is
	 * exactly equal to <code>degree</code>.
	 * @throws IllegalArgumentException if <code>degree</code>
	 * is not (strictly) positive.
	 */
	public ComplexSet getRoots(int degree) {
		if (degree <= 0)
			throw new IllegalArgumentException("degree must be strictly positive");
		
		ComplexSet roots = new ComplexSet(degree);
		
		double thisPhase = this.phase();
		// absolute value of each root 
		// is equal to the degree-th root of
		// this.abs()
		double rootAbs = Math.pow(this.abs(), 1.0d/degree);
		
		// phases
		for (int k = 0; k < degree; k++) {
			double rootPhase = (thisPhase + (2.0d * k * Math.PI)) /
					(degree * 1.0d);
			roots.add(Complex.fromPolarForm(rootAbs, rootPhase));
		}
		
		return roots;
	}
	
}
