package it.polimi.deib.se.numbers;

import java.util.Arrays;


/**
 * Represents a set of complex numbers with a
 * maximum size.
 * 
 * The set mutable and it is implemented such that 
 * two complex numbers are considered equal if their
 * difference in absolute value is less than the specified
 * precision, which is an immutable parameter of the
 * set. The implementation of this set depends on the
 * specification of {@link Complex#approximatelyEquals(Complex, double)}.
 * 
 * 
 * @see Complex#approximatelyEquals(Complex, double)
 * @author Andrea Mocci
 * @version 1.0
 * @since 1.0
 */
public class ComplexSet {

	// Rappresentazione interna: array di elementi
	private final Complex[] elements;
	// dimensione effettiva dell'insieme
	private int size;
	// precisione per controllare l'uguaglianza
	private final double precision;
	// costante che rappresenta la dimensione massima del set
	private final int maxSize; 
	

	/**
	 * Builds a set of complex numbers with a 
	 * default precision and the specified maximum size.
	 * @param maxSize the maximum size of the set.
	 * @throws IllegalArgumentException if <code>maxSize</code> is
	 * negative.
	 */
	public ComplexSet(int maxSize) {
		this(maxSize, 2.0d * Double.MIN_VALUE);		
	}
	
	/**
	 * Builds a set of complex numbers with a specified
	 * precision for equality check and a specified
	 * maximum size.
	 * @param maxSize the maximum size of the set.
	 * @param precision the precision for equality check.
	 * @throws IllegalArgumentException if <code>maxSize</code> is
	 * negative.
	 * @throws IllegalArgumentException if <code>precision</code>
	 * is either negative, infinite or NaN.
	 */
	public ComplexSet(int maxSize, double precision) {
		if (maxSize < 0)
			throw new IllegalArgumentException("...");
		this.size = 0;
		this.maxSize = maxSize;
		this.elements = new Complex[maxSize];
		this.precision = precision;
	}

	/**
	 * Adds a new element to the set.
	 * 
	 * The element is added to the set only if the set
	 * is not full. Moreover, the element is added only if
	 * is not already present and no other element is 
	 * approximately equal to it considering the
	 * precision of this set. 
	 * @param element a complex number to add.
	 * @throws IllegalArgumentException if <code>element</code>
	 * is <code>null</code>.
	 * @throws FullSetException if the set is full.
	 * @see ComplexSet#isFull()
	 */
	public void add(Complex element) {
		if (element == null)
			throw new IllegalArgumentException("element cannot be null");
		
		// invoco il metodo che controlla 
		// la presenza dell'elemento nel set.
		 boolean containsElement = this.contains(element);
		 
		 // se e' gia' presente non faccio nulla
		 if (containsElement)
			 return;
		 
		 /* controllo che il set non sia pieno */
		 if (this.size < this.maxSize) {
			 this.elements[size] = element;
			 this.size++;		 
		 } /* se il set e' pieno, non faccio nulla */
		 else {
			 throw new FullSetException();
		 }
	}
	
	/**
	 * Checks if the set is full.
	 * 
	 * If the set is full, the {@link #add(Complex)}
	 * operation is disabled and behaves exceptionally.
     *
	 * @return <code>true</code> if the set is full,
	 * <code>false</code> otherwise.
	 * @see #add(Complex)
	 */
	public boolean isFull() {
		return this.size == this.maxSize;
	}
	
	/**
	 * Checks if the set contains an element.
	 * 
	 * Coherently with the semantics of this class, 
	 * the containment check is performed according
	 * to the approximate equality of complex numbers.
	 * 
	 * @param element a non-null complex number.
	 * @return <code>true</code> if the set contains the
	 * specified element, <code>false</code> otherwise.
	 * @throws IllegalArgumentException if <code>element</code>
	 * is <code>null</code>.
	 * @see Complex#approximatelyEquals(Complex, double)
	 */
	public boolean contains(Complex element) {
		for (int i = 0; i < this.size; i++) {
			if (elements[i].approximatelyEquals(element, this.precision))
				return true;
		}

		return false;
	}
		
	/**
	 * Removes an element from the set.
	 * 
	 * Coherently with the semantics of this class, 
	 * the search is performed according
	 * to the approximate equality of complex numbers.
	 * 
	 * @param element a non-null complex number.
	 * @throws IllegalArgumentException if <code>element</code>
	 * is <code>null</code>.
	 * @see Complex#approximatelyEquals(Complex, double)
	 */
	public void remove(Complex element) {
		// recupero l'elemento
		int pos = find(element);
		
		// shift elementi: evito di lasciare un buco
		// nell'array
		if (pos >= 0) {
			for(int i = pos + 1; i < this.size - 1; i++)
				elements[i] = elements[i + 1];
			this.size = this.size - 1;
		}

	}
	
	// NB: Ho documentato il metodo privato con javadoc solo
	// per completezza, e per evidenziare che la specifica
	// di questo metodo dipende (e non puo' non dipendere)
	// dalla rappresentazione interna,
	// quindi il metodo deve essere private.
	/**
	 * Tries to search for a complex number in the set,
	 * and returns its position in the internal array if
	 * it is present
	 * 
	 * @param element a non-null complex number
	 * @return the position of <code>element</code> in the
	 * <code>elements</code> array, or <code>-1</code> if
	 * the element is not presetnt.
	 */
	private int find(Complex element) {
		for (int i = 0; i < this.size; i++) {
			if (elements[i].approximatelyEquals(element, this.precision))
				return i;
		}
		/* default */
		return -1;
	}
		
	// Quando vedremo un'astrazione sul controllo opportuna,
	// (gli iteratori), questo metodo avra' meno senso.
	/**
	 * Returns an array of {@link Complex} containing
	 * the elements of the set.
	 * @return an array of {@link Complex} containing
	 * the elements of the set.
	 */
	public Complex[] toArray() {
		return Arrays.copyOf(this.elements, this.size);
	}
	
	/**
	 * Returns the actual size of the set.
	 * @return an integer representing the size of the set,
	 * that is, the number of elements contained.
	 */
	public int size() {
		return this.size;
	}	
	
	/**
	 * Returns the maximum possible size of the set.
	 * 
	 * @return an integer representing the maximum possible
	 * size of the set.
	 */
	public int getMaxSize() {
		return this.maxSize;
	}

}


